@@ -12,6 +12,7 @@
 
 #include <vicious.h>
 
+#include "gdmwm.h"
 #include "greeter_configuration.h"
 #include "greeter_parser.h"
 #include "greeter_events.h"
@@ -19,20 +20,47 @@
 /* FIXME: hack */
 extern GreeterItemInfo *welcome_string_info;
 
+/* evil globals */
 static char *file_search_path = NULL;
+static GList *button_stack = NULL;
+
+static GHashTable *pixbuf_hash = NULL;
 
 GHashTable *item_hash = NULL;
+GList *custom_items = NULL;
 
 static gboolean parse_items (xmlNodePtr       node,
 			     GList          **items_out,
 			     GreeterItemInfo *parent,
 			     GError         **error);
 
+static GdkPixbuf *
+load_pixbuf (const char *fname, GError **error)
+{
+  GdkPixbuf *pb;
+
+  if (pixbuf_hash == NULL)
+    pixbuf_hash = g_hash_table_new_full (g_str_hash,
+					 g_str_equal,
+					 g_free,
+					 (GDestroyNotify)g_object_unref);
+  pb = g_hash_table_lookup (pixbuf_hash, fname);
+  if (pb != NULL)
+    return g_object_ref (pb);
+
+  pb = gdk_pixbuf_new_from_file (fname, error);
+  if G_UNLIKELY (pb == NULL)
+    return NULL;
+
+  g_hash_table_insert (pixbuf_hash, g_strdup (fname), g_object_ref (pb));
+
+  return pb;
+}
 
 GQuark
 greeter_parser_error_quark (void)
 {
-  static GQuark quark;
+  static GQuark quark = 0;
   if (!quark)
     quark = g_quark_from_static_string ("greeter_parser_error");
 
@@ -152,7 +180,7 @@ parse_pos (xmlNodePtr       node,
   if (prop)
     {
       info->x = g_ascii_strtod (prop, &p);
-      
+
       if ((char *)prop == p)
 	{
 	  g_set_error (error,
@@ -162,6 +190,11 @@ parse_pos (xmlNodePtr       node,
 	  xmlFree (prop);
 	  return FALSE;
 	}
+
+      if (prop[0] == '-' || info->x < 0)
+        info->x_negative = TRUE;
+      else
+        info->x_negative = FALSE;
       
       if (strchr (prop, '%') != NULL)
 	info->x_type = GREETER_ITEM_POS_RELATIVE;
@@ -184,6 +217,11 @@ parse_pos (xmlNodePtr       node,
 	  xmlFree (prop);
 	  return FALSE;
 	}
+
+      if (prop[0] == '-' || info->y < 0)
+        info->y_negative = TRUE;
+      else
+        info->y_negative = FALSE;
       
       if (strchr (prop, '%') != NULL)
 	info->y_type = GREETER_ITEM_POS_RELATIVE;
@@ -292,7 +330,7 @@ parse_show (xmlNodePtr       node,
 
   /* Note: subtype is deprecated, use type only */
   prop = xmlGetProp (node, "subtype");
-  if (prop != NULL)
+  if G_UNLIKELY (prop != NULL)
     {
       /* code for legacy uses of subtype only, are there any such
        * themes out there?  The Bluecurve was the one this was made
@@ -439,7 +477,7 @@ parse_box (xmlNodePtr       node,
     {
       info->box_x_padding = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
@@ -456,7 +494,7 @@ parse_box (xmlNodePtr       node,
     {
       info->box_y_padding = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
@@ -473,7 +511,7 @@ parse_box (xmlNodePtr       node,
     {
       info->box_min_width = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
@@ -490,7 +528,7 @@ parse_box (xmlNodePtr       node,
     {
       info->box_min_height = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
@@ -507,7 +545,7 @@ parse_box (xmlNodePtr       node,
     {
       info->box_spacing = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
@@ -534,7 +572,7 @@ parse_color (const char *str,
 {
   guint32 col;
   int i;
-  if (str[0] != '#')
+  if G_UNLIKELY (str[0] != '#')
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
@@ -542,7 +580,7 @@ parse_color (const char *str,
 		   "colors must start with #, %s is an invalid color\n", str);
       return FALSE;
     }
-  if (strlen (str) != 7)
+  if G_UNLIKELY (strlen (str) != 7)
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
@@ -569,6 +607,8 @@ parse_state_file (xmlNodePtr node,
 {
   xmlChar *prop;
   char *p;
+
+  info->have_state |= (1<<state);
   
   prop = xmlGetProp (node, "file");
   if (prop)
@@ -588,7 +628,7 @@ parse_state_file (xmlNodePtr node,
     {
       if (!parse_color (prop, &info->tints[state], error))
 	return FALSE;
-      info->have_tint[state] = TRUE;
+      info->have_tint |= (1<<state);
       xmlFree (prop);
     }
 
@@ -597,7 +637,7 @@ parse_state_file (xmlNodePtr node,
     {
       info->alphas[state] = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
@@ -620,13 +660,15 @@ parse_state_color (xmlNodePtr node,
 {
   xmlChar *prop;
   char *p;
+
+  info->have_state |= (1<<state);
   
   prop = xmlGetProp (node, "color");
   if (prop)
     {
-      if (!parse_color (prop, &info->colors[state], error))
+      if G_UNLIKELY (!parse_color (prop, &info->colors[state], error))
 	return FALSE;
-      info->have_color[state] = TRUE;
+      info->have_color |= (1<<state);
       xmlFree (prop);
     }
 
@@ -635,7 +677,7 @@ parse_state_color (xmlNodePtr node,
     {
       info->alphas[state] = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
@@ -665,49 +707,49 @@ parse_pixmap (xmlNodePtr        node,
     {
       if (strcmp (child->name, "normal") == 0)
 	{
-	  if (!parse_state_file (child, info, GREETER_ITEM_STATE_NORMAL, error))
+	  if G_UNLIKELY (!parse_state_file (child, info, GREETER_ITEM_STATE_NORMAL, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "prelight") == 0)
 	{
-	  if (!parse_state_file (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
+	  if G_UNLIKELY (!parse_state_file (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "active") == 0)
 	{
-	  if (!parse_state_file (child, info, GREETER_ITEM_STATE_ACTIVE, error))
+	  if G_UNLIKELY (!parse_state_file (child, info, GREETER_ITEM_STATE_ACTIVE, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "pos") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_pos (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0)
 	{
-	  if (!parse_fixed (child, info, error))
+	  if G_UNLIKELY (!parse_fixed (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "box") == 0)
 	{
-	  if (!parse_box (child, info, error))
+	  if G_UNLIKELY (!parse_box (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
 	    return FALSE;
 	}
       
       child = child->next;
     }
 
-  if (!info->files[GREETER_ITEM_STATE_NORMAL])
+  if G_UNLIKELY (!info->files[GREETER_ITEM_STATE_NORMAL])
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
 		   GREETER_PARSER_ERROR_BAD_SPEC,
-		   "No filename specified for normal state\n");
+		   "No filename specified for normal state");
       return FALSE;
     }
   
@@ -717,13 +759,13 @@ parse_pixmap (xmlNodePtr        node,
 	{
 	  if (info->files[i] != NULL)
 	    {
-	      info->orig_pixbufs[i] = gdk_pixbuf_new_from_file (info->files[i], error);
+	      info->pixbufs[i] = load_pixbuf (info->files[i], error);
 	      
-	      if (info->orig_pixbufs[i] == NULL)
+	      if G_UNLIKELY (info->pixbufs[i] == NULL)
 		return FALSE;
 	    }
 	  else
-	    info->orig_pixbufs[i] = NULL;
+	    info->pixbufs[i] = NULL;
 	}
     }
 
@@ -744,37 +786,37 @@ parse_rect (xmlNodePtr node,
     {
       if (strcmp (child->name, "normal") == 0)
 	{
-	  if (!parse_state_color (child, info, GREETER_ITEM_STATE_NORMAL, error))
+	  if G_UNLIKELY (!parse_state_color (child, info, GREETER_ITEM_STATE_NORMAL, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "prelight") == 0)
 	{
-	  if (!parse_state_color (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
+	  if G_UNLIKELY (!parse_state_color (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "active") == 0)
 	{
-	  if (!parse_state_color (child, info, GREETER_ITEM_STATE_ACTIVE, error))
+	  if G_UNLIKELY (!parse_state_color (child, info, GREETER_ITEM_STATE_ACTIVE, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "pos") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_pos (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0)
 	{
-	  if (!parse_fixed (child, info, error))
+	  if G_UNLIKELY (!parse_fixed (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "box") == 0)
 	{
-	  if (!parse_box (child, info, error))
+	  if G_UNLIKELY (!parse_box (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
 	    return FALSE;
 	}
       
@@ -783,7 +825,7 @@ parse_rect (xmlNodePtr node,
 
   for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
     {
-      if (!info->have_color[i])
+      if ( ! (info->have_color & (1<<i)))
 	continue;
       
       if (info->alphas[i] >= 1.0)
@@ -807,11 +849,13 @@ parse_state_text (xmlNodePtr node,
   xmlChar *prop;
   char *p;
 
+  info->have_state |= (1<<state);
+
   prop = xmlGetProp (node, "font");
   if (prop)
     {
       info->fonts[state] = pango_font_description_from_string (prop);
-      if (info->fonts[state] == NULL)
+      if G_UNLIKELY (info->fonts[state] == NULL)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
@@ -826,9 +870,9 @@ parse_state_text (xmlNodePtr node,
   prop = xmlGetProp (node, "color");
   if (prop)
     {
-      if (!parse_color (prop, &info->colors[state], error))
+      if G_UNLIKELY (!parse_color (prop, &info->colors[state], error))
 	return FALSE;
-      info->have_color[state] = TRUE;
+      info->have_color |= (1<<state);
       xmlFree (prop);
    }
 
@@ -837,7 +881,7 @@ parse_state_text (xmlNodePtr node,
     {
       info->alphas[state] = g_ascii_strtod (prop, &p);
       
-      if ((char *)prop == p)
+      if G_UNLIKELY ((char *)prop == p)
 	{
 	  g_set_error (error,
 		       GREETER_PARSER_ERROR,
@@ -951,6 +995,21 @@ parse_stock (xmlNodePtr node,
 	  g_free (*translated_text);
 	  *translated_text = g_strdup (_("_Quit"));
 	}
+      else if (g_ascii_strcasecmp (prop, "halt") == 0)
+        {
+	  g_free (*translated_text);
+	  *translated_text = g_strdup (_("Shut_down"));
+	}
+      else if (g_ascii_strcasecmp (prop, "suspend") == 0)
+        {
+	  g_free (*translated_text);
+	  *translated_text = g_strdup (_("Sus_pend"));
+	}
+      else if (g_ascii_strcasecmp (prop, "reboot") == 0)
+        {
+	  g_free (*translated_text);
+	  *translated_text = g_strdup (_("_Reboot"));
+	}
       else if (g_ascii_strcasecmp (prop, "chooser") == 0)
         {
 	  g_free (*translated_text);
@@ -1019,6 +1078,7 @@ parse_label (xmlNodePtr        node,
 {
   xmlNodePtr child;
   int i;
+  double size_reduction = 1.0;
   char *translated_text;
   gint translation_score = 1000;
   
@@ -1029,34 +1089,39 @@ parse_label (xmlNodePtr        node,
     {
       if (strcmp (child->name, "normal") == 0)
 	{
-	  if (!parse_state_text (child, info, GREETER_ITEM_STATE_NORMAL, error))
+	  if G_UNLIKELY (!parse_state_text (child, info, GREETER_ITEM_STATE_NORMAL, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "prelight") == 0)
 	{
-	  if (!parse_state_text (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
+	  if G_UNLIKELY (!parse_state_text (child, info, GREETER_ITEM_STATE_PRELIGHT, error))
+	    return FALSE;
+	}
+      else if (strcmp (child->name, "active") == 0)
+	{
+	  if G_UNLIKELY (!parse_state_text (child, info, GREETER_ITEM_STATE_ACTIVE, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "pos") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_pos (child, info, error))
 	    return FALSE;
 	}
       else if (child->type == XML_ELEMENT_NODE &&
 	       strcmp (child->name, "text") == 0)
 	{
-	  if (!parse_translated_text (child, &translated_text, &translation_score, error))
+	  if G_UNLIKELY (!parse_translated_text (child, &translated_text, &translation_score, error))
 	    return FALSE;
 	}
       else if (child->type == XML_ELEMENT_NODE &&
 	       strcmp (child->name, "stock") == 0)
 	{
-	  if (!parse_stock (child, info, &translated_text, &translation_score, error))
+	  if G_UNLIKELY (!parse_stock (child, info, &translated_text, &translation_score, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0 ||
@@ -1078,6 +1143,7 @@ parse_label (xmlNodePtr        node,
 		   GREETER_PARSER_ERROR,
 		   GREETER_PARSER_ERROR_BAD_SPEC,
 		   "A label must specify the text attribute");
+      return FALSE;
     }
   /* FIXME: evil hack to use internally translated strings */
   if (translation_score == 999 &&
@@ -1090,7 +1156,7 @@ parse_label (xmlNodePtr        node,
 
   for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
     {
-      if (!info->have_color[i])
+      if ( ! (info->have_color & (1<<i)))
 	continue;
       
       if (info->alphas[i] >= 1.0)
@@ -1104,11 +1170,91 @@ parse_label (xmlNodePtr        node,
   if (info->fonts[GREETER_ITEM_STATE_NORMAL] == NULL)
     info->fonts[GREETER_ITEM_STATE_NORMAL] = pango_font_description_from_string ("Sans");
 
+  if (gdm_wm_screen.width <= 800 &&
+      gdm_wm_screen.width > 640)
+    size_reduction = PANGO_SCALE_SMALL;
+  else if (gdm_wm_screen.width <= 640)
+    size_reduction = PANGO_SCALE_X_SMALL;
+
+  if (size_reduction < 0.99)
+    {
+      for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
+        {
+          if (info->fonts[i] != NULL)
+	    {
+	      int old_size = pango_font_description_get_size (info->fonts[i]);
+	      pango_font_description_set_size (info->fonts[i], old_size * size_reduction);
+	    }
+	}
+    }
+
   info->orig_text = translated_text;
   
   return TRUE;
 }
 
+static gboolean
+parse_listitem (xmlNodePtr        node,
+		GreeterItemInfo  *info,
+		GError         **error)
+{
+  xmlNodePtr child;
+  char *translated_text = NULL;
+  gint translation_score = 1000;
+  xmlChar *prop;
+  GreeterItemListItem *li;
+  
+  prop = xmlGetProp (node, "id");
+  
+  if G_LIKELY (prop)
+    {
+      li = g_new0 (GreeterItemListItem, 1);
+      li->id = g_strdup (prop);
+      xmlFree (prop);
+    }
+  else
+    {
+      g_set_error (error,
+		   GREETER_PARSER_ERROR,
+		   GREETER_PARSER_ERROR_BAD_SPEC,
+		   "Listitem id not specified");
+      return FALSE;
+    }
+
+  child = node->children;
+  while (child)
+    {
+      if (child->type == XML_ELEMENT_NODE &&
+	  strcmp (child->name, "text") == 0)
+	{
+	  if G_UNLIKELY ( ! parse_translated_text (child, &translated_text, &translation_score, error))
+	    {
+              g_free (li->id);
+              g_free (li);
+	      return FALSE;
+	    }
+	}
+    
+      child = child->next;
+    }
+
+  if G_UNLIKELY (translated_text == NULL)
+    {
+      g_free (li->id);
+      g_free (li);
+      g_set_error (error,
+		   GREETER_PARSER_ERROR,
+		   GREETER_PARSER_ERROR_BAD_SPEC,
+		   "A list item must specify the text attribute");
+      return FALSE;
+    }
+  li->text = translated_text;
+
+  info->list_items = g_list_append (info->list_items, li);
+
+  return TRUE;
+}
+
 static gboolean
 parse_list (xmlNodePtr        node,
 	     GreeterItemInfo  *info,
@@ -1121,12 +1267,17 @@ parse_list (xmlNodePtr        node,
     {
       if (strcmp (child->name, "pos") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_pos (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
+	    return FALSE;
+	}
+      else if (strcmp (child->name, "listitem") == 0)
+	{
+	  if G_UNLIKELY ( ! parse_listitem (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0 ||
@@ -1142,6 +1293,17 @@ parse_list (xmlNodePtr        node,
       child = child->next;
     }
 
+  if (info->list_items != NULL) {
+    if G_UNLIKELY (strcmp (info->id, "userlist") == 0) {
+      g_set_error (error,
+		   GREETER_PARSER_ERROR,
+		   GREETER_PARSER_ERROR_BAD_SPEC,
+		   "List of id userlist cannot have custom list items");
+      return FALSE;
+    }
+    custom_items = g_list_append (custom_items, info);
+  }
+
   return TRUE;
 }
 
@@ -1157,12 +1319,12 @@ parse_entry (xmlNodePtr        node,
     {
       if (strcmp (child->name, "pos") == 0)
 	{
-	  if (!parse_pos (child, info, error))
+	  if G_UNLIKELY (!parse_pos (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "show") == 0)
 	{
-	  if (!parse_show (child, info, error))
+	  if G_UNLIKELY (!parse_show (child, info, error))
 	    return FALSE;
 	}
       else if (strcmp (child->name, "fixed") == 0 ||
@@ -1193,6 +1355,8 @@ parse_items (xmlNodePtr  node,
     xmlChar *type;
     GreeterItemInfo *info;
     GreeterItemType item_type;
+
+    *items_out = NULL;
     
     items = NULL;
     
@@ -1201,7 +1365,7 @@ parse_items (xmlNodePtr  node,
       {
 	if (child->type == XML_ELEMENT_NODE)
 	  {
-	    if (strcmp (child->name, "item") != 0)
+	    if G_UNLIKELY (strcmp (child->name, "item") != 0)
 	      {
 		g_set_error (error,
 			     GREETER_PARSER_ERROR,
@@ -1211,7 +1375,7 @@ parse_items (xmlNodePtr  node,
 	      }
 	    
 	    type = xmlGetProp (child, "type");
-	    if (!type)
+	    if G_UNLIKELY (!type)
 	      {
 		g_set_error (error,
 			     GREETER_PARSER_ERROR,
@@ -1245,9 +1409,14 @@ parse_items (xmlNodePtr  node,
 	    info = greeter_item_info_new (parent, item_type);
 	    
 	    parse_id (child, info);
-	    if ( ! parse_button (child, info, error))
+	    if G_UNLIKELY ( ! parse_button (child, info, error))
 	      return FALSE;
 
+	    if (button_stack != NULL)
+	      info->my_button = button_stack->data;
+	    if (info->button)
+	      button_stack = g_list_prepend (button_stack, info);
+
 	    switch (item_type)
 	      {
 	      case GREETER_ITEM_TYPE_SVG:
@@ -1275,8 +1444,11 @@ parse_items (xmlNodePtr  node,
 			     "Bad item type");
 		res = FALSE;
 	      }
+
+	    if (info->button)
+	      button_stack = g_list_remove (button_stack, info);
 	    
-	    if (!res)
+	    if G_UNLIKELY (!res)
 	      return FALSE;
 
 	    items = g_list_prepend (items, info);
@@ -1314,9 +1486,10 @@ greeter_parse (const char *file, const char *datadir,
   GList *items;
 
   /* FIXME: EVIL! GLOBAL! */
+  g_free (file_search_path);
   file_search_path = g_strdup (datadir);
   
-  if (!g_file_test (file, G_FILE_TEST_EXISTS))
+  if G_UNLIKELY (!g_file_test (file, G_FILE_TEST_EXISTS))
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
@@ -1327,7 +1500,7 @@ greeter_parse (const char *file, const char *datadir,
   
 
   doc = xmlParseFile (file);
-  if (doc == NULL)
+  if G_UNLIKELY (doc == NULL)
     {
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
@@ -1337,8 +1510,9 @@ greeter_parse (const char *file, const char *datadir,
     }
   
   node = xmlDocGetRootElement (doc);
-  if (node == NULL)
+  if G_UNLIKELY (node == NULL)
     {
+      xmlFreeDoc (doc);
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
 		   GREETER_PARSER_ERROR_BAD_XML,
@@ -1346,8 +1520,9 @@ greeter_parse (const char *file, const char *datadir,
       return NULL;
     }
   
-  if (strcmp (node->name, "greeter") != 0)
+  if G_UNLIKELY (strcmp (node->name, "greeter") != 0)
     {
+      xmlFreeDoc (doc);
       g_set_error (error,
 		   GREETER_PARSER_ERROR,
 		   GREETER_PARSER_ERROR_WRONG_TYPE,
@@ -1356,21 +1531,43 @@ greeter_parse (const char *file, const char *datadir,
     }
 
 
-  item_hash = g_hash_table_new_full ((GHashFunc)greeter_info_id_hash,
-				     (GEqualFunc)greeter_info_id_equal,
-				     NULL,
-				     (GDestroyNotify)greeter_item_info_free);
+  item_hash = g_hash_table_new ((GHashFunc)greeter_info_id_hash,
+				(GEqualFunc)greeter_info_id_equal);
   
 
   root = greeter_item_info_new (NULL, GREETER_ITEM_TYPE_RECT);
   res = parse_items (node, &items, root, error);
-  if (!res)
+
+  /* Now we can whack the hash, we don't want to keep cached
+     pixbufs around anymore */
+  g_hash_table_destroy (pixbuf_hash);
+  pixbuf_hash = NULL;
+
+  if G_UNLIKELY (!res)
     {
+      welcome_string_info = NULL;
+
+      g_hash_table_destroy (item_hash);
+      item_hash = NULL;
+      g_list_free (custom_items);
+      custom_items = NULL;
+
+      g_list_free (button_stack);
+      button_stack = NULL;
+
+      g_list_foreach (items, (GFunc) greeter_item_info_free, NULL);
+      g_list_free (items);
+      items = NULL;
+
       greeter_item_info_free (root);
 
+      xmlFreeDoc (doc);
+
       return NULL;
     }
 
+  xmlFreeDoc (doc);
+
   root->fixed_children = items;
   
   root->x = 0;
@@ -1387,3 +1584,9 @@ greeter_parse (const char *file, const char *datadir,
   
   return root;
 }
+
+const GList *
+greeter_custom_items (void)
+{
+  return custom_items;
+}
